# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/nanoGPT_P2_PyPI/04_viz.ipynb.

# %% auto 0
__all__ = ['read_log', 'plot_log']

# %% ../nbs/nanoGPT_P2_PyPI/04_viz.ipynb 6
import matplotlib.pyplot as plt
import numpy as np

# %% ../nbs/nanoGPT_P2_PyPI/04_viz.ipynb 10
def read_log(log_file="log/log.txt"):
    """read the log file"""

    # load the log file
    with open(log_file, "r") as f:
        lines = f.readlines()
    
    # parse the individual lines, group by stream (train,val,hella)
    streams = {}
    for line in lines:
        step, stream, val = line.strip().split()
        if stream not in streams:
            streams[stream] = {}
        streams[stream][int(step)] = float(val)

    # convert each stream from {step: val} to (steps[], vals[])
    # so it's easier for plotting
    streams_xy = {}
    for k, v in streams.items():
        # get all (step, val) items, sort them
        xy = sorted(list(v.items()))
        # unpack the list of tuples to tuple of lists
        streams_xy[k] = list(zip(*xy))

    return streams_xy

# %% ../nbs/nanoGPT_P2_PyPI/04_viz.ipynb 12
def plot_log(log_file="log/log.txt", sz="124M"):
    """plot the log file"""

    loss_baseline = {"124M": 3.2924,}[sz]
    # HellaSwag for GPT-2
    hella2_baseline = {"124M": 0.294463, "350M": 0.375224, "774M": 0.431986, "1558M": 0.488946,}[sz]
    # HellaSwag for GPT-3
    hella3_baseline = {"124M": 0.337, "350M": 0.436, "774M": 0.510, "1558M": 0.547,}[sz]

    streams_xy = read_log(log_file=log_file)
    # create figure
    plt.figure(figsize=(16, 6))
    # Panel 1: losses: both train and val
    plt.subplot(121)
    xs, ys = streams_xy["train"] # training loss
    ys = np.array(ys)
    plt.plot(xs, ys, label=f'nanogpt ({sz}) train loss')
    print("Min Train Loss:", min(ys))
    xs, ys = streams_xy["val"] # validation loss
    plt.plot(xs, ys, label=f'nanogpt ({sz}) val loss')
    # horizontal line at GPT-2 baseline
    if loss_baseline is not None:
        plt.axhline(y=loss_baseline, color='r', linestyle='--', label=f"OpenAI GPT-2 ({sz}) checkpoint val loss")
    plt.xlabel("steps"); plt.ylabel("loss")
    plt.yscale('log'); plt.ylim(2.0, 12.0)
    plt.legend(); plt.title("Loss")
    print("Min Validation Loss:", min(ys))

    # Panel 2: HellaSwag eval
    plt.subplot(122)
    xs, ys = streams_xy["hella"] # HellaSwag eval
    ys = np.array(ys)
    plt.plot(xs, ys, label=f"nanogpt ({sz})")
    # horizontal line at GPT-2 baseline
    if hella2_baseline:
        plt.axhline(y=hella2_baseline, color='r', linestyle='--', label=f"OpenAI GPT-2 ({sz}) checkpoint")
    if hella3_baseline:
        plt.axhline(y=hella3_baseline, color='g', linestyle='--', label=f"OpenAI GPT-3 ({sz}) checkpoint")
    plt.xlabel("steps"); plt.ylabel("accuracy")
    plt.legend(); plt.title("HellaSwag eval")
    print("Max Hellaswag eval:", max(ys))
